package mixer;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Dictionary;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;

import com.google.protobuf.InvalidProtocolBufferException;

import javafx.util.Pair;

import java.util.Random;

import meerkat.protobuf.Crypto.*;
import meerkat.protobuf.Mixing.*;

import meerkat.crypto.Encryption;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;

public class Mixer{
	
	private Random random;
	private Mix2ZeroKnowledgeProver prover; 
	private Encryption encryptor;
	
	public Mixer(Random rand,Mix2ZeroKnowledgeProver prov,Encryption enc) {
		this.random = rand;
		this.prover = prov;
		this.encryptor = enc;
		
	}
	
	public Pair<ZeroKnowledgeProof[][],RerandomizableEncryptedMessage[][]> mix(List<RerandomizableEncryptedMessage> ciphertexts) throws InvalidProtocolBufferException{
		
		int n = ciphertexts.size();
		// assert n = 2^k and n > 1
		if( n <= 1 || ((n & (n-1)) != 0))
			return null;

		//initialization
		int layers = 0;
		for (int i = n; i > 1; i>>=1) {
			layers++;
		}
        layers<<=1;
        layers--;
        RerandomizableEncryptedMessage[][] encryptionsTable = new RerandomizableEncryptedMessage[layers][n];
    	ZeroKnowledgeProof[][] proofsTable= new ZeroKnowledgeProof[layers][n/2];
		boolean[][] mixnet = createMixNet(n,layers);
        int index1, index2, switchIndex = 0;
        EncryptionRandomness r1 ,r2;
        RerandomizableEncryptedMessage e1, e2;
        boolean half = true;
        
        //set first level of encryptions
        for (int j = 0; j < n; j++)
        {
            encryptionsTable[0][j] = ciphertexts.get(j);
        }

        // main loop 
        for (int i = n, layer = 0; layer < layers; layer++) // i == permutation size
        {
            for (int j = 0; j < n; j += i) // 
            {
                for (int k = 0; k < i / 2; k++)
                {
                    index1 = k + j;
                    index2 = k + j + i / 2;
                    e1 = encryptionsTable[layer][index1];
                    e2 = encryptionsTable[layer][index2];
                    r1 = encryptor.generateRandomness(random);
                    r2 = encryptor.generateRandomness(random);
                    if (!mixnet[layer][switchIndex])
                    {
                        encryptionsTable[layer+1][index1] = encryptor.rerandomize(e1, r1);
                        encryptionsTable[layer+1][index2] = encryptor.rerandomize(e2,r2);
                        
                    }
                    else
                    {
                    	encryptionsTable[layer+1][index1] = encryptor.rerandomize(e2,r2);
                        encryptionsTable[layer+1][index2] = encryptor.rerandomize(e1,r1);
                    }
                    proofsTable[layer][switchIndex] =
                    		prover.prove(e1, e2, encryptionsTable[layer + 1][index1], 
                    				encryptionsTable[layer + 1][index2],
                    				 mixnet[layer][switchIndex], r1,r2);
                    
                    switchIndex = (switchIndex + 1) % (n / 2);
                }
            }
            if (half)
            {
                i >>= 1;
                if (i == 1)
                {
                    half = false;
                    i = 4;
                }
            }
            else
            {
                i <<= 1;
            }
        }
        return  new Pair<ZeroKnowledgeProof[][],RerandomizableEncryptedMessage[][]>(proofsTable, encryptionsTable);
	}
	
	private int[] randomPermutation(int n){
		List<Integer> numbers= new ArrayList<Integer>(n);
		for (int i = 0; i < n; i++)
		{
			numbers.add(i);
		}

		int[] result = new int[n];
		int index;
		for (int i = 0; i < n; i++)
		{
			index = random.nextInt(n - i);
			result[i] = numbers.get(index);
			numbers.remove(index);
		}
		return result;
	}
	
	private boolean[][] createMixNet(int n,int layers)
	{
		int[] permutaion = randomPermutation(n);
		int[] pi, piL, piR;
		Queue<int[]> permutaions = new ArrayBlockingQueue<int[]>(n);
		Graph graph;
		boolean[][] mixnet = new boolean[layers][n>>1];

		permutaions.add(permutaion);

		for (int i = n, layer = 0; i > 1; i >>= 1, layer++) // i == permutation size
				{
			for (int j = 0; j < n / 2; j += i / 2) // 
			{
				pi = permutaions.remove();
				graph = new Graph(pi);
				piL = new int[i / 2];
				piR = new int[i / 2];
				for (int k = 0; k < i / 2; k++)
				{
					mixnet[layers - layer - 1][k + j] = graph.getSwitchValue(k, true);
					mixnet[layer][k + j] = graph.getSwitchValue(k, false);

					if (!mixnet[layers - layer - 1][k + j])
					{
						piL[k] = pi[k] % (i / 2);
						piR[k] = pi[k + i / 2] % (i / 2);
					}
					else
					{
						piL[k] = pi[k + i / 2] % (i / 2);
						piR[k] = pi[k] % (i / 2);
					}
				}
				permutaions.add(piL);
				permutaions.add(piR);
			}
				}
		return mixnet;
	}

}
